import { protocolIsRegistered } from "./protocol";
export class ChainContext {
    network;
    platform;
    chain;
    config;
    // Cached Protocol clients
    rpc;
    tokenBridge;
    autoTokenBridge;
    circleBridge;
    autoCircleBridge;
    ibcBridge;
    constructor(chain, platform) {
        this.config = platform.config[chain];
        this.platform = platform;
        this.chain = this.config.key;
        this.network = this.config.network;
    }
    getRpc() {
        this.rpc = this.rpc ? this.rpc : this.platform.getRpc(this.chain);
        return this.rpc;
    }
    // Get the number of decimals for a token
    async getDecimals(token) {
        return this.platform.utils().getDecimals(this.chain, this.getRpc(), token);
    }
    // Get the balance of a token for a given address
    async getBalance(walletAddr, token) {
        return this.platform.utils().getBalance(this.chain, await this.getRpc(), walletAddr, token);
    }
    async getLatestBlock() {
        return this.platform.utils().getLatestBlock(this.getRpc());
    }
    async getLatestFinalizedBlock() {
        return this.platform.utils().getLatestFinalizedBlock(this.getRpc());
    }
    // Get details about the transaction
    async parseTransaction(txid) {
        return this.platform.parseWormholeMessages(this.chain, await this.getRpc(), txid);
    }
    // Send a transaction and wait for it to be confirmed
    async sendWait(stxns) {
        return this.platform.utils().sendWait(this.chain, await this.getRpc(), stxns);
    }
    async getNativeWrappedTokenId() {
        const tb = await this.getTokenBridge();
        return { chain: this.chain, address: await tb.getWrappedNative() };
    }
    // Get the token account for a given address
    async getTokenAccount(address, token) {
        // Noop by default, override in implementation if necessary
        return { chain: this.chain, address };
    }
    //
    // protocols
    //
    //
    supportsProtocol(protocolName) {
        return protocolIsRegistered(this.chain, protocolName);
    }
    supportsTokenBridge = () => this.supportsProtocol("TokenBridge");
    async getTokenBridge() {
        this.tokenBridge = this.tokenBridge
            ? this.tokenBridge
            : await this.platform.getProtocol("TokenBridge", await this.getRpc());
        return this.tokenBridge;
    }
    //
    supportsAutomaticTokenBridge = () => this.supportsProtocol("AutomaticTokenBridge");
    async getAutomaticTokenBridge() {
        this.autoTokenBridge = this.autoTokenBridge
            ? this.autoTokenBridge
            : await this.platform.getProtocol("AutomaticTokenBridge", await this.getRpc());
        return this.autoTokenBridge;
    }
    //
    supportsCircleBridge = () => this.supportsProtocol("CircleBridge");
    async getCircleBridge() {
        this.circleBridge = this.circleBridge
            ? this.circleBridge
            : await this.platform.getProtocol("CircleBridge", await this.getRpc());
        return this.circleBridge;
    }
    //
    supportsAutomaticCircleBridge = () => this.supportsProtocol("AutomaticCircleBridge");
    async getAutomaticCircleBridge() {
        this.autoCircleBridge = this.autoCircleBridge
            ? this.autoCircleBridge
            : await this.platform.getProtocol("AutomaticCircleBridge", await this.getRpc());
        return this.autoCircleBridge;
    }
    //
    supportsIbcBridge = () => this.supportsProtocol("IbcBridge");
    async getIbcBridge() {
        this.ibcBridge = this.ibcBridge
            ? this.ibcBridge
            : await this.platform.getProtocol("IbcBridge", await this.getRpc());
        return this.ibcBridge;
    }
}
//# sourceMappingURL=chain.js.map